<h3>Useful Links</h3>
<ul>
  <li><a href="https://pypi.python.org/pypi/{{project}}">{{project}} @ PyPI</a></li>
  <li><a href="https://github.com/jlesquembre/{{project}}">{{project}} @ github</a></li>
  <li><a href="https://github.com/jlesquembre/{{project}}/issues">report an error</a></li>
</ul>
